/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <time.h>

#include <gssapi.h>

#include "llgt_config.h"

#include "llgt_globus_internal.h"

#include "llgt_utils.h"

static int llgt_debugging_mode = 0; /* disabled by default */

void llgt_enable_debugging_mode(void) {
    llgt_debugging_mode = 1;
}

int llgt_is_debugmode_enabled(void) {
    return llgt_debugging_mode;
}




void llgt_logmsg (int priority, const char* format, ...)
{
    va_list     args;

    /* Check if the module was enabled with debugging enabled.
     * If so, pass the LOG_DEBUG level message though */
    if (priority == LOG_DEBUG && !llgt_is_debugmode_enabled()) {
        /* Skip debugging messages */
        return;
    }

    va_start(args, format);
    vsyslog(priority, format, args);
    va_end(args);
}



char* llgt_get_client_name(gss_ctx_id_t context)
{
	int initiator;
	int major_status;
	unsigned int minor_status;
	gss_name_t peer;
	gss_buffer_desc peer_name_buffer;
	char *client_name = NULL;

	major_status = gss_inquire_context(&minor_status,
                                       context,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       &initiator,
                                       GLOBUS_NULL);
    if(GSS_ERROR(major_status))
    {
        /* GLOBUS_GRIDMAP_CALLOUT_GSS_ERROR(result, major_status, minor_status); */
        /* goto error; */
    	return NULL;
    }

	major_status = gss_inquire_context(&minor_status,
                                       context,
                                       initiator ? GLOBUS_NULL : &peer,
                                       initiator ? &peer : GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL,
                                       GLOBUS_NULL);
    if(GSS_ERROR(major_status))
    {
		return NULL;
    }
    major_status = gss_display_name(&minor_status,
                                    peer,
                                    &peer_name_buffer,
                                    GLOBUS_NULL);
    if(GSS_ERROR(major_status))
    {
        /* GLOBUS_GRIDMAP_CALLOUT_GSS_ERROR(result, major_status, minor_status); */
        gss_release_name(&minor_status, &peer);
		return NULL;
    }
    gss_release_name(&minor_status, &peer);
	if((client_name = (char*)malloc(peer_name_buffer.length + 1)))
	{
		memcpy(client_name, peer_name_buffer.value, peer_name_buffer.length);
		client_name[peer_name_buffer.length] = 0;
	}
	gss_release_buffer(&minor_status, &peer_name_buffer);
	return client_name;
}



gss_cred_id_t llgt_get_user_cred_handle(gss_ctx_id_t context_handle)
{
    llgt_gss_ctx_id_desc *local_handle=(llgt_gss_ctx_id_desc*)context_handle;
    return local_handle->peer_cred_handle;
}



int llgt_create_jobid(void)
{
    /*
     * Create the unique Job Repository ID and export it as an environment variable
     * the ID is the same one as for the Job Manager with the addition of microseconds.
     * We moved this part here so that the LCMAPS Job Repository plugin
     * can use this environment variable to uniquely identify the job
     * To have a really unique id we also add the microseconds to the time string
     */
    struct tm  * jr_tmp;
    const char * jr_id_var = "JOB_REPOSITORY_ID";
    char         jr_id[71];
    const char * gk_jm_id_var = "GATEKEEPER_JM_ID";
    char         gatekeeper_jm_id[64];
    const pid_t  gatekeeper_pid = getpid();
    static int   reqnr = 0;

#ifdef HAVE_GETTIMEOFDAY
    {
    struct timeval  jr_clock;

    /* get time in microseconds ! */
    gettimeofday(&jr_clock,NULL);
    jr_tmp = gmtime(&(jr_clock.tv_sec));

    snprintf(jr_id, 70, "%04d-%02d-%02d.%02d:%02d:%02d.%06d.%010u.%010u",
        jr_tmp->tm_year + 1900, jr_tmp->tm_mon + 1, jr_tmp->tm_mday,
        jr_tmp->tm_hour, jr_tmp->tm_min, jr_tmp->tm_sec, (int)(jr_clock.tv_usec),
        gatekeeper_pid & 0xFFFFFFFF, reqnr & 0xFFFFFFFF);
    }
#else
    {
    time_t       myclock;

    time(&myclock);
    jr_tmp = gmtime(&myclock);
    }

    snprintf(jr_id, 70, "%04d-%02d-%02d.%02d:%02d:%02d.%010u.%010u",
        jr_tmp->tm_year + 1900, jr_tmp->tm_mon + 1, jr_tmp->tm_mday,
        jr_tmp->tm_hour, jr_tmp->tm_min, jr_tmp->tm_sec,
        gatekeeper_pid & 0xFFFFFFFF, reqnr & 0xFFFFFFFF);
#endif
    snprintf(gatekeeper_jm_id, 63, "%04d-%02d-%02d.%02d:%02d:%02d.%010u.%010u",
        jr_tmp->tm_year + 1900, jr_tmp->tm_mon + 1, jr_tmp->tm_mday,
        jr_tmp->tm_hour, jr_tmp->tm_min, jr_tmp->tm_sec,
        gatekeeper_pid & 0xFFFFFFFF, reqnr & 0xFFFFFFFF);


    setenv(jr_id_var, jr_id, 1);
    setenv(gk_jm_id_var, gatekeeper_jm_id, 1);

    reqnr++;
    return 0;
}
